import numpy
from numpy import linalg


class Matrix:
    """
    This Matrix class represents a real 2D-matrix.
    """
    __data__: numpy.ndarray
    __shape__: (int, int)

    def __init__(self, data=None, shape=None, structure=None, model=None, n=None):
        """
        Creates a new matrix.
        Thy type of the matrix depends on the signature and arguments.

        - ``Matrix(numpy.ndarray)``: will create a new matrix with the given data in ndarray and its shape.
        - ``Matrix(list, (int,int))``: will create a new nxm matrix with the given rows and columns and data in list.
        - ``Matrix(list, str, int)``: will create a new square matrix of given size and structure of either \"unity\", \"diagonal\" or \"tridiagonal\"
        - ``Matrix(str, int)``: will create a new square matrix of given size and TODO


        :param data: Either a list or an numpy ndarray
        :param shape: A tuple containing the amount of rows and columns
        :param structure: Either \"unity\", \"diagonal\" or \"tridiagonal\"
        :param model: TODO
        :param n: Amount of rows of a square matrix or offset in case of diagonal structure

        :type data: list | numpy.ndarray
        :type shape: (int, int)
        :type structure: str
        :type model: str
        :type n: int

        :rtype: Matrix
        """
        # Case Matrix(str, int)
        if n is not None and model is not None:
            ...  # TODO: what shall one do here?
        # Case: Matrix(list, str, int)
        elif n is not None and structure is not None and data is not None:
            if structure == "unity":
                ...  # TODO: what does it mean?
            elif structure == "diagonal":
                diag = numpy.diag(data, n)
                self.__data__ = diag
                self.__shape__ = diag.shape
            elif structure == "tridiagonal":
                if len(data) != 3:
                    raise ValueError("If structure is tridiagonal, then the given data must be of length 3")
                tridiag = numpy.diag([data[0]] * (n - 1), -1) + numpy.diag([data[1]] * n, 0) + numpy.diag(
                    [data[2]] * (n - 1), 1)
                self.__data__ = tridiag
                self.__shape__ = tridiag.shape
        # Case: Matrix(list, str, int)
        elif shape is not None and data is not None:
            self.__shape__ = shape
            self.__data__ = numpy.array(data).reshape(shape)
        # Case: Matrix(numpy.ndarray)
        elif data is not None:
            try:
                data.shape[1]
            except IndexError:
                self.__shape__ = (data.shape[0], 1)
            else:
                self.__shape__ = data.shape
            self.__data__ = data
        else:
            raise ValueError(
                "Only following signatures are allowed: (numpy.ndarray), (list, tuple), (list, str, int), (str, int)")

    def get_data(self):
        """
        :return: the data of the matrix as a ``numpy.ndarray``
        """
        return self.__data__

    def shape(self):
        """
        :return: the shape of the matrix, which is ``(rows, columns)``
        """
        return self.__shape__

    def transpose(self):
        """
        :return: the transpose of the matrix
        """
        return Matrix(self.__data__.transpose())

    def __eq__(self, other):
        """
        Return ``self==value``

        :param other: The object to compare to; must be either a ``Matrix``, a ``list`` or a ``numpy.ndarray``
        :return: True if data in the matrix are equal to the given data in other for each component, otherwise False
        """
        if isinstance(other, Matrix):
            to_compare = other.__data__
        elif isinstance(other, list):
            to_compare = numpy.array(other)
        elif isinstance(other, numpy.ndarray):
            to_compare = other
        else:
            raise ValueError("Matrix type is not comparable to type of given ``other``")
        return (self.__data__ == to_compare).all()

    def __str__(self):
        return str(self.__data__)

    def __neg__(self):
        return Matrix(-self.__data__)

    def __add__(self, other):
        if isinstance(other, Matrix):
            if self.__shape__ != other.__shape__:
                raise ValueError("The shape of the operands must be the same")
            return Matrix(self.__data__ + other.__data__)
        elif isinstance(other, int) or isinstance(other, float):
            return Matrix(self.__data__ + other)
        else:
            raise ValueError("Only a number or another ``Matrix`` can be added to a ``Matrix``")

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return -self + other

    def __mul__(self, other):
        if isinstance(other, Matrix):
            if self.__shape__[1] != other.__shape__[0]:
                raise ValueError(
                    "The amount of columns of the first operand must match the amount of rows of the second operand")
            return Matrix(self.__data__ @ other.__data__)
        elif isinstance(other, int) or isinstance(other, float):
            return Matrix(other * self.__data__)
        else:
            raise ValueError("Only a number or another ``Matrix`` can be multiplied to a ``Matrix``")

    def __rmul__(self, other):
        return self * other

    def __truediv__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            return Matrix(self.__data__ / other)
        else:
            raise ValueError("A ``Matrix`` can only be divided ba a number")

    def norm(self, f: str = "frobenius"):
        """
        Calculates the norm of the matrix.

        A norm is a positive definit, absolute homogeneous and subadditive function.
        For Matrices a norm is also sub-multiplicative.

        :param f: The norm to be used, could be either "frobenius", "rowsum" or "colsum"

        :return: the norm as a number
        """
        t = "fro"
        if f == "colsum":
            t = 1
        elif f == "rowsum":
            t = numpy.inf
        return linalg.norm(self.__data__, t)

    def __getitem__(self, key):
        return self.__data__[key]

    def __setitem__(self, key, value):
        self.__data__[key] = value
